//#define TIMER
//#define TEST_COVAR_JP
#define TEST_COVAR_BM

#include <mex.h>
#include "covSum.h"

/*
 * This is the static version of the covSum.m function. Currently it calcualtes 
 * covariance matrix as a sum of covSEard and covNoise functions. Additionaly,
 * it returns derivatives along hyperparameters.
 *
 * Input parameters:
 *   loghyper - hyperparameters
 *   D - size of loghyper
 *   X - input data
 *   N - X length
 *   z - hyperparameter coordinate, iff 0 then calculate cov. matirx
 *
 * The function uses static matrices defined in the ggpr.h header
 *
 */
void covSum( real *loghyper, int D, int M, size_t NinDim, size_t Mtd, int z )
{
  real s2;
  real sf2;
  real *ell;

  dim3 dimBlock( NBLOCKSX, NBLOCKSY );
  dim3 dimGrid( ( M + dimBlock.x - 1 )/dimBlock.x,	\
		( M + dimBlock.y - 1 )/dimBlock.y	\
		);
  
#ifdef TIMER
  mexPrintf("    begin covSum()\n");
#endif
  checkStatusCUDA("covSum1 ");
  TIMER_INIT
  checkStatusCUDA("covSum2 ");

#ifdef LOGINFO_COV
  mexPrintf( "covSum Entry\n" );
  mexPrintf ( "input_GPU = 0x%08x\n", input_GPU );	
  mexPrintf ( "K         = 0x%08x\n", K );	
#endif

#ifdef DEBUG_VAR_DUMP
  {
	real *host_X = NULL;					/* Host memory block. */
	size_t host_X_width = M*sizeof(real);	/* Width in bytes of a row in matrix X. */

	/* Allocate the host memory for a copy of K matrix. */
	cudaMallocHost ( &host_X, NinDim * host_X_width );
	/* Copy the X from the device to the host. */
	cudaMemcpy2D ( host_X, host_X_width, input_GPU, pitchIn, host_X_width, NinDim, cudaMemcpyDeviceToHost );
	/* Print out the contents of X. */
    mexPrintf ( "X_device[%d,%d]=\n", M, NinDim );
	for ( int i = 0; i < NinDim ; i++ )
	{
	  for ( int j = 0; j < M ; j++ )
	  {
		mexPrintf ( "  %6.4f", host_X[M*i+j] );
	  }
	  mexPrintf ( "\n" );
	}
	/* Free the page-locked host memory block. */
	cudaFreeHost ( host_X );
  }
#endif

  /* check D */
  s2  = exp( 2 * loghyper[D-1] ); /* covNoise parameter */
  sf2 = exp( 2 * loghyper[D-2] ); /* covSEard parameters */
#ifdef LOGINFO_COV
  mexPrintf( "s2 = %f\n", s2 );
  mexPrintf( "sf2 = %f\n", sf2 );
#endif

  /* malloc*/
  ell = (real *) malloc( sizeof( real ) * ( D - 2 ) );
  for ( int i = 0; i < D-2; i++ ) 
  {
    ell[i] = exp( loghyper[i] ); /* problem */
#ifdef LOGINFO_COV
    mexPrintf( "ell[%d] = %f\n", i, ell[i] );
#endif
  }

#ifdef LOGINFO_COV
  mexPrintf("copy ell to device\n");
#endif
  checkStatusCUDA("covSum3a ");
  cudaMemcpy( gell, ell, sizeof( real ) * ( D - 2 ), cudaMemcpyHostToDevice );
  checkStatusCUDA("covSum3b ");

  if ( z == 0 )                                                        /* compute covariance matrix */
  { 
    /* call kernel to initialize matrix */
    TIMER_START;
#ifdef LOGINFO_COV
    mexPrintf ( "  -- z=0, calcCovarMatrix()\n" );
    mexPrintf ( "     M=%d, NinDim=%d, D=%d\n" , M, NinDim, D );
#endif

#ifdef TEST_COVAR_JP
      TIMER_START;

      dimBlock.x = 16;
	  dimBlock.y = 16;
	  dimGrid.x = ( M + dimBlock.x - 1 )/dimBlock.x;
	  dimGrid.y = ( M + dimBlock.y - 1 )/dimBlock.y;
      size_t shmSize = dimBlock.x*dimBlock.y*sizeof(real);

#ifdef DEBUG
      mexPrintf ( "    - M = %d\n", M );
	  mexPrintf ( "      NinDim = %d\n", NinDim );
      mexPrintf ( "      block is %dx%dx%d threads\n", dimBlock.x, dimBlock.y, dimBlock.z );
      mexPrintf ( "      grid is %dx%dx%d blocks\n", dimGrid.x, dimGrid.y, dimGrid.z );
      mexPrintf ( "      dynamic shared memory %d bytes\n", shmSize );
#endif
#ifdef NEVER
      /* calculate Kstar */
	  calcTestCovarMatrixJP<<<dimGrid,dimBlock,shmSize>>>(
        K, pitch,
		input_GPU, pitchIn,
        M,
		input_GPU, pitchIn,
        M,
		s2, sf2, gell,
		NinDim
		);
	  TIMER_STOP("calcTestCovarMatrixJP")
#endif
#ifdef DEBUG
      {
        size_t nr = (M>5)?5:M;
        size_t nc = (Mtd>5)?5:Mtd;
        dumpVariable ( "K_JP", Kstar, pitchKstar, nr, nc );
      }
#endif
#endif // TEST_COVAR_JP
  }
  else  /* compute derivative along z dimension of hyperparameter vector */
  {
    if ( z > 0 && ( z <= NinDim )) /* should be: z > 0 && z <= (int) NinDim */
	{
#ifdef LOGINFO_COV
      mexPrintf ( "  -- z=%d, calcDerMatrix()\n", z );
      mexPrintf ( "     M=%d, NinDim=%d, D=%d\n" , M, NinDim, D );
#endif
      checkStatusCUDA("ts1 ");
	  TIMER_START;
      checkStatusCUDA("ts2 ");
	  calcCovarMatrixSEARD<<<dimGrid,dimBlock>>>( K, pitch,		\
						      input_GPU, pitchIn, \
						      M, NinDim, sf2, gell \
						      );
      checkStatusCUDA("ts3 ");
      TIMER_STOP("calcCovarMatrixSEARD");
      checkStatusCUDA("ts4 ");

	  /* check and update K matrix --> not implemented */
	  TIMER_START;
      checkStatusCUDA("ts5 ");
	  calcDerMatrix<<<dimGrid,dimBlock>>>( Der, pitchDer,	\
					       K, pitch,	   \
					       input_GPU, pitchIn, \
					       M, z-1, gell	   \
					       );
      checkStatusCUDA("ts6 ");
      TIMER_STOP("calcDerMatrix");
      checkStatusCUDA("ts7 ");
	}
    else if ( z == D - 1 )
	{
   	  TIMER_START;
	  /* check and update K matrix --> not implemented */
	  /*	      calcCovarMatrixSEARD<<<dimGrid,dimBlock>>>(K, pitch, input_GPU, pitchIn, M, z, sf2, ell);*/ 
	  /* A = 2*K */
	  timesTwoCopy<<<dimGrid,dimBlock>>>( Der, pitchDer, K, pitch, M );
	  /* clear K --> not implemented*/
      TIMER_STOP("timesTwoCopy");
	}
    else if ( z == D )
	{
	  /* first set to 0 */
	  TIMER_START;
	  cudaMemset2D( Der, pitchDer, 0, M * sizeof( real ), M );
      TIMER_STOP("memset Der to 0");
	  /* init diagonal elements */
	  TIMER_START;
      calcDerMatrixNoise<<<dimGrid,dimBlock>>>( Der, pitchDer, M, s2 );
      TIMER_STOP("calcDerMatrixNoise");
	}
    else if ( z == -1 )                                           /* compute test set covariances */
	{
	  TIMER_START;
	  /* setup partitioning for Kss */
	  dim3 dimBlockV( BLOCKSIZE );
	  dim3 dimGridV( ( Mtd + dimBlockV.x - 1 ) / dimBlockV.x );
	  calcKss<<<dimGridV,dimBlockV>>>( Kss, Mtd, sf2, s2 );
	  TIMER_STOP("calcKss");

#ifdef DEBUG
      {
        size_t nr = (M>5)?5:M;
        size_t nc = (NinDim>5)?5:NinDim;
        dumpVariable ( "x", input_GPU, pitchIn, nr, nc );
      }

      {
        size_t nr = (Mtd>5)?5:Mtd;
        size_t nc = (NinDim>5)?5:NinDim;
        dumpVariable ( "xstar", inputStar, pitchInStar, nr, nc );
      }
#endif

#ifdef TEST_COVAR_JP
      TIMER_START;

      dimBlock.x = 16;
	  dimBlock.y = 16;
	  dimGrid.x = ( Mtd + dimBlock.x - 1 )/dimBlock.x;
	  dimGrid.y = ( M + dimBlock.y - 1 )/dimBlock.y;
      size_t shmSize = dimBlock.x*dimBlock.y*sizeof(real);

#ifdef DEBUG
      mexPrintf ( "    - Mtd = %d\n", Mtd );
	  mexPrintf ( "      M = %d\n", M );
	  mexPrintf ( "      NinDim = %d\n", NinDim );
      mexPrintf ( "      block is %dx%dx%d threads\n", dimBlock.x, dimBlock.y, dimBlock.z );
      mexPrintf ( "      grid is %dx%dx%d blocks\n", dimGrid.x, dimGrid.y, dimGrid.z );
      mexPrintf ( "      dynamic shared memory %d bytes\n", shmSize );
#endif
      /* calculate Kstar */
	  calcTestCovarMatrixJP<<<dimGrid,dimBlock,shmSize>>>(
        Kstar, pitchKstar,
		input_GPU, pitchIn,
        M,
		inputStar, pitchInStar,
        M,
		s2, sf2, gell,
		NinDim
		);
	  TIMER_STOP("calcTestCovarMatrixJP")
#ifdef DEBUG
      {
        size_t nr = (M>5)?5:M;
        size_t nc = (Mtd>5)?5:Mtd;
        dumpVariable ( "KstarJP", Kstar, pitchKstar, nr, nc );
      }
#endif
#endif // TEST_COVAR_JP

#ifdef TEST_COVAR_BM
      TIMER_START;
	  
      dimBlock.x = 16;
	  dimBlock.y = 16;
	  dimGrid.x = ( Mtd + dimBlock.x - 1 )/dimBlock.x;
	  dimGrid.y = ( M + dimBlock.y - 1 )/dimBlock.y;

#ifdef DEBUG
      mexPrintf ( "    - Mtd = %d\n", Mtd );
	  mexPrintf ( "      M = %d\n", M );
	  mexPrintf ( "      NinDim = %d\n", NinDim );
      mexPrintf ( "      block is %dx%dx%d threads\n", dimBlock.x, dimBlock.y, dimBlock.z );
      mexPrintf ( "      grid is %dx%dx%d blocks\n", dimGrid.x, dimGrid.y, dimGrid.z );
#endif

      /* calculate Kstar */
	  calcTestCovarMatrixBM<<<dimGrid,dimBlock>>>( Kstar, pitchKstar, \
						     input_GPU, pitchIn, \
						     inputStar, pitchInStar, \
						     Mtd, M,		\
						     s2, sf2, gell,	\
						     NinDim		\
						     );
	  TIMER_STOP("calcTestCovarMatrixBM")
#ifdef DEBUG
      {
        size_t nr = (M>5)?5:M;
        size_t nc = (Mtd>5)?5:Mtd;
        dumpVariable ( "KstarBM", Kstar, pitchKstar, nr, nc );
      }
#endif
#endif // TEST_COVAR_BM
	}
  }
#ifdef LOGINFO_COV
  mexPrintf("free ell\n");
#endif
  free( ell );
#ifdef TIMER
  mexPrintf("    end covSum()\n");
#endif
}

/* Dump a submatrix d_var(1:nr,1:nc) */
void dumpVariable (
  const char *name,
  double *d_var,
  size_t  pitch,
  size_t  nr,
  size_t  nc
  )
{
  double * h_var;
  unsigned int rid,cid;
  /* Allocate nx*pich bytes of memory and copy data from the device to
     the host memory. */
  cudaMallocHost ( (void**)&h_var, nc * pitch );
  /* Copy the X from the device to the host. */
  cudaMemcpy2D ( (void*)h_var, pitch, (void*)d_var, pitch, pitch, nc, cudaMemcpyDeviceToHost );
  mexPrintf ( "    %s (first %d rows and %d columns):\n", name, nr, nc );
  for ( rid=0 ; rid<nr ; rid++ )
  {
    mexPrintf ( "     " );
    for ( cid=0 ; cid<nc ; cid++ )
    {
      double *col = (double*)((char*)h_var + cid*pitch );
      mexPrintf ( " % 10.6f", col[rid] );
    }
    mexPrintf ( "\n" );
  }
  cudaFreeHost((void*)h_var);
}