/* This is needed for MSVC compiler in order to define M_PI in math.h */
#define _USE_MATH_DEFINES

//#define TIMER

#include <math.h>
#include "mex.h"
#include "ggpr.h"
#include "covSum.h"

/* mex gpr - Gaussian process regression 

Call this from Matlab as

[m,s] = ggpr ( logtheta, covfunc, x, y, xstar )

where

logtheta .... a (column) vector of log hyperparameters
covfunc ..... ignored
x ........... a n by D matrix of training inputs
y ........... a (column) vector (of size n) of targets
xstar ....... a nn by D matrix of test inputs

Below, x => input, y => target, xstar => testdata.

*/ 

/* Initialisation flags. */
int initialized      = 0; 
int initializedTD    = 0;
int initializedTest  = 0; /* alpha=K^-1*y and L^-1 comptued and cached */
int cuda_device      = 0; /* can be selected only in case that the card is not in use */

/* GPU variables */
real  *input_GPU = NULL;   /* input data on device */
size_t pitchIn;
size_t NinDim;
real  *target_GPU = NULL;
real  *K = NULL;
size_t pitch;
real  *L = NULL;
size_t pitchL;
real  *Der = NULL;         /* derivatives matrix */
size_t pitchDer;
real  *W = NULL;           /* matrix for intermediate results */
size_t pitchW;
real  *inputStar = NULL;
size_t pitchInStar;
real  *Kstar = NULL;       /* Kstar matrix in gpr.m */
size_t pitchKstar;

real *alpha = NULL; /* solution */
real *copyAlpha = NULL; /* copy of Alpha */
real *diag = NULL;  /* diagonal elements of L */
real *outDiag = NULL;
real *Kss = NULL;
real *gout1 = NULL;        /* buffer for output parameter Kss */
real *gell = NULL;       

real *gsum;         /* temporary colum sums for sum(sum()) implemented as reduction */

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

  /* host input variables */
  real *input_host;            /* ptr to host input data */
  real *target_host;           /* ptr to host target data*/
  real *testdata_host;
  real *logtheta; 

  real *ISum;  /* interbediate sum */

  /* local variables */
  real *out1;                      /* output parameters */
  real *out2;
#ifdef OUTTEST
  real *out3;
#endif
  int M;                          /* size of input data */
  int Mtd;                         /* size of test data */
  int N;          /* number of columns in input vectors */
  int D;                   /* number of hyperparameters */
  int NinDim;             /* number of input dimensions */

  cublasStatus cublasStat;              /* cublas status*/

  dim3 dimBlock;
  dim3 dimGrid;


  TIMER_INIT


  /* check input parameters */

#ifdef OUTTEST
  /* Third output parameter is used to output some selected internal structure. */
  if (nlhs > 2)
#else
  if (nlhs > 1)
#endif
  mexErrMsgTxt("Wrong number of output parameters.");

  if (nrhs != 3)
    mexErrMsgTxt("Wrong number of input parameters.");

  /* ignore first parameter - covariance function*/

  logtheta = (real *) mxGetPr( prhs[1] ); 
  D = mxGetM( prhs[1] );
  N = mxGetN( prhs[1] );
  
  if ( N > D )
    mexErrMsgTxt("Input vector logtheta should be row oriented.");

  input_host = (real *) mxGetPr(prhs[2]);

  M = mxGetM(prhs[2]);
  NinDim = mxGetN(prhs[2]);

  if ( NinDim == 0 && M == 0 ) {              /* check if reset was requested */
    gprcleanup();  /* reset memory */    
#ifdef DEBUG
    mexPrintf ( "cleanup finished, returning\n");
#endif
    return;
  }

  if ( NinDim > M )
    mexErrMsgTxt("Input vector x should be row oriented.");

  if ( cuda_device == -1 )
  {
    checkStatusCUDA("select1 ");
  }
#ifdef DEBUG
  cudaDeviceProp properties;
  
  checkStatusCUDA("select2 ");
  cudaGetDeviceProperties ( &properties, cuda_device );
  checkStatusCUDA("select3 ");

  mexPrintf (
    "\n"
    "--------------------\n"
    "Using Device[%d]: %s, %d processors, %d global memory, %d shm/block\n"
    "--------------------\n",
    cuda_device,
    properties.name,
    properties.multiProcessorCount,
    properties.totalGlobalMem,
    properties.sharedMemPerBlock);
#endif  
  
  
  /* First check if memory has been initialized ... */
  if ( !initialized )
  {
    /* ... and if not, allocate memory and copy data to the device. */
    TIMER_START

#ifdef DEBUG 
    mexPrintf ( "ggpr() initialisation\n" );
    mexPrintf ( "  M = %d\n", M );
    mexPrintf ( "  NinDim = %d\n", NinDim );
    mexPrintf ( "  D = %d\n", D );
#endif

    /* Memory allocation for all global variables. Returns false in case
       that there was some error during the memory allocation process
       (this would be probably "out of memory" error). */
    if ( gprinit( M, NinDim, D ))
    {
      /* Memory allocated, now copy data to the device. */
      cudaMemcpy2D (
        input_GPU, pitchIn,
        input_host, sizeof(real)*M, sizeof(real)*M, NinDim,
        cudaMemcpyHostToDevice );                /* copy input data to GPU */

      mexAtExit(gprcleanup);                     /* register cleanup function */
    }
    else
    {
      mexErrMsgTxt ( "gprinit() failed.");
    }
    TIMER_STOP("  initialisation")
    checkStatusCUDA ("  initialisation ");
  }

  
  /* compute training set covariance matrix */
  TIMER_START;
  covSum( logtheta, D,
	    M, NinDim, Mtd,
	    0
	    );
  TIMER_STOP ( "covSum" );
  

      /* allocate output covariance matrix */
      TIMER_START;
	  plhs[0] = mxCreateDoubleMatrix( M, M, mxREAL );
	  out2 = mxGetPr( plhs[0] ); /* variances */
	  cudaMemcpy2D( out2, sizeof(double)*M, K, pitch, sizeof(double)*M, M, cudaMemcpyDeviceToHost);
      checkStatusCUDA("diffsumsq3 ");
	  TIMER_STOP ( "    memcopy cov to host");

#ifdef TIMER
  mexPrintf ( "-----\n" );
#endif
}

int gprinit( int M, int NinDim, int D )
{
  cudaError_t error;
  int haveError = 0;

#ifdef DEBUG
  mexPrintf (
    "gprinit(%d,%d,%d): allocating %d bytes of memory\n",
    M, NinDim, D,
    (unsigned long)(M*sizeof(real)*NinDim) + 
    (unsigned long)(4*sizeof(real)*M) + 
    (unsigned long)(4*M*sizeof(real)*M) + 
    sizeof( real ) * ( D - 2 ) );
#endif

  error = cudaMallocPitch( (void **) &input_GPU, &pitchIn, M*sizeof(real), NinDim );
  haveError |= printErrorMsg ( "  input_GPU ", error );
#ifdef DEBUG
  mexPrintf ( "  input_GPU is %d bytes\n", NinDim*pitchIn );
#endif

  error = cudaMallocPitch( (void **) &K, &pitch, M*sizeof(real), M );
  haveError |= printErrorMsg ( "  K ", error );
#ifdef DEBUG
  mexPrintf ( "  K is %d bytes\n", pitch*M );
#endif

  error = cudaMalloc( (void **) &gell, sizeof( real ) * ( D - 2 ) );
  haveError |= printErrorMsg ( "  gell", error );

  //  cudaMalloc( (void **) &diag, 2*sizeof(real)*dimBlock.x*dimGrid.x); /* pitched vector*/
  //  cudaMalloc( (void **) &outDiag, sizeof(real)*dimGrid.x);

  if ( ! haveError )
  {
    initialized = 1; /* initialization done */
  }

  return ( ! haveError );
}

void cudaFreeAllocated ( real **data )
{
  if ( *data != NULL )
  {
    cudaFree ( *data );
    *data = NULL;
  }

}

void gprcleanup(void)
{
  if ( initialized )
  {
#ifdef DEBUG
    mexPrintf( "MEX-file is terminating, destroying arrays\n" );
#endif
    cudaFreeAllocated ( &K );
    cudaFreeAllocated ( &L );
    cudaFreeAllocated ( &Der );
    cudaFreeAllocated ( &W );
    cudaFreeAllocated ( &input_GPU );
    cudaFreeAllocated ( &target_GPU );
    cudaFreeAllocated ( &alpha );
    cudaFreeAllocated ( &copyAlpha );
    //    cudaFree( diag );0
    cudaFreeAllocated ( &gell );
    cudaFreeAllocated ( &gsum );
    initialized = 0;
    initializedTest = 0;
  }
#ifdef DEBUG
  else
  {
    mexPrintf( "gprcleanup(): main data not initialised, skipping.\n" );
  }
#endif

  if ( initializedTD )
  {
#ifdef DEBUG
    mexPrintf( "Destroying inputStar arary\n" );
#endif
    cudaFreeAllocated ( &inputStar );
    cudaFreeAllocated ( &Kstar );
    cudaFreeAllocated ( &Kss );
    cudaFreeAllocated ( &gout1 );
    initializedTD = 0;
  }
#ifdef DEBUG
  else
  {
    mexPrintf( "gprcleanup(): test data not initialised, skipping.\n" );
  }
#endif

  cudaThreadExit();
  checkStatusCUDA ( "gprcleanup(): " );
#ifdef DEBUG
    mexPrintf( "gprcleanup() finished.\n" );
#endif
}


void checkStatusCUBLAS(cublasStatus status) {
  if ( status != CUBLAS_STATUS_SUCCESS)
    {

      mexPrintf("cublasInit error\n");
    }
}

void checkStatusCUDA ( const char *msg )
{
  cudaError_t e = cudaThreadSynchronize();
  if ( e != cudaSuccess )
  {
    mexPrintf ( "%sCUDA threadsynch error (%d): %s\n", msg, (int)e, cudaGetErrorString(e) );
  }
  else
  {
    e = cudaGetLastError();
    printErrorMsg ( msg, e );
  }
}

int printErrorMsg ( const char *msg, cudaError_t e )
{
  int res = 0;
  if ( e != cudaSuccess )
  {
    mexPrintf ( "%sCUDA error (%d): %s\n", msg, (int)e, cudaGetErrorString(e) );
    res = 1;
  }
  return res;
}