#ifndef __GGPR_H__
#define __GGPR_H__

#include "driver_types.h"
#include "cuda.h"
#include "cublas.h"

#define real double

#ifdef TIMER
#define TIMER_INIT  cudaEvent_t start, stop; \
                    float elapsedTime; \
                    cudaEventCreate(&start); \
		    cudaEventCreate(&stop);
#define TIMER_START   cudaEventRecord(start,0);
#define TIMER_STOP(cmt)   cudaEventRecord(stop,0); \
                          cudaEventSynchronize(stop); \
                          cudaEventElapsedTime(&elapsedTime, start, stop); \
			  mexPrintf( cmt  " : t_el = %f\n",elapsedTime);

#else
#define TIMER_INIT
#define TIMER_START
#define TIMER_STOP(cmt)
#endif

#define BLOCKSIZE 512
#define NBLOCKSX 16
#define NBLOCKSY 16


extern int initialized; 
extern int initializedTD;

/* GPU variables are global, initialised just once as the MAX file gets
   called for the first time. */
extern  real *input_GPU;   /* input data on device */
extern  size_t pitchIn;
extern  size_t NinDim;
extern  real *target_GPU;
extern  real *K;
extern  size_t pitch;
extern  real *L;
extern  size_t pitchL;
extern  real *Der;         /* derivatives matrix */
extern  size_t pitchDer;
extern  real *W;           /* matrix for intermediate results */
extern  size_t pitchW;
extern  real *inputStar;
extern  size_t pitchInStar;
extern  real *Kstar;       /* Kstar matrix in gpr.m */
extern  size_t pitchKstar;


extern  real *alpha; /* solution */
extern  real *diag;  /* diagonal elements of L */
extern  real *outDiag;
extern  real *Kss;
extern  real *gout1;        /* buffer for output parameter Kss */
extern  real *gell;         

/* 
 * function for global memory allocation
 */
int gprinit( int M, int NinDim, int D ); 

/*
 *  cleans static memory on device
 */
void gprcleanup( void );

/*
 * utility function for displaying CUBLAS messages
 */
void checkStatusCUBLAS( cublasStatus status );

void checkStatusCUDA ( const char *msg );
int printErrorMsg ( const char *msg, cudaError_t e );


/* kernels */

/* 
 *  kernel copies diagonal elements form matrix S to vector D
 */
__global__ void copyVecDiag(real* D, real* S, size_t pitch, int M);

/* 
 * kernel used to copy two vectors
 */
__global__ void copyVec(real* D, real* S, size_t M);

/*
 * parallel prefix sum (based on CUDA example)
 */
__global__ void reduce0(real *g_idata, real *g_odata);

/*
 * parallel prefix sum (based on CUDA example). For matrices
 */
__global__ void reduce0M(real *g_idata, real *g_odata);

/*
 * init matrix set 1 on diagonal
 */
__global__ void eye(real *A, size_t pitch, size_t M);

/*
 * Kernel calculatec v = v.*v;
 */
__global__ void  MatMul(real *A, size_t pitchA, real *B, size_t pitchB, size_t M);

/* 
 *  Initialization of Kss vector 
 */
__global__ void calcKss( real *Kss, size_t Nin, real sf2, real s2);

/* 
 * This kernel creates the covar. matrix  (covSEard + covNoise)
 */
__global__ void calcTestCovarMatrixBM( real *Kstar, size_t pitchKstar,	\
				     real *X, size_t pitchx,		\
				     real *Xstar, size_t pitchXstar,	\
				     size_t N, size_t M,		\
				     real s2,				\
				     real sf2,				\
				     real *ell,				\
				     size_t NinDim			\
				     );

__global__ void calcTestCovarMatrixJP
(
  real   *Kstar, size_t pitchKstar,
  real   *X,     size_t pitchX,
  size_t  M,
  real   *Xstar, size_t pitchXstar,
  size_t  Mtd,
  real    s2,
  real    sf2,
  real   *ell,
  size_t  NinDim
  );

/*
 * Kernel calculates v = v.*v;
 */ 
__global__ void  squareElements( real *C, size_t pitchC, size_t M, size_t N );


template <class T>
void gpusum_wrapper (
    int size,
    int threads,
    int blocks,
    T *d_data
    );

template <class T>
void dotwder_wrapper (
    int size,
    int threads,
    int blocksX,
    int blocksY, 
    T *d_iwdata,
    T *d_iddata,
    size_t pitch,
    T *d_odata
    );

template <class T>
void diffvsummsq_wrapper (
    int sizex,
    int sizey,
    int threads,
    int blocksX,
    int blocksY, 
    T *d_ivdata,
    T *d_imdata,
    size_t pitch,
    T *d_odata
    );

void get_gpusum_blkthr (
  int n,
  int maxBlocks,
  int maxThreads,
  int &blocks,
  int &threads
  );

void get_diffvsummsq_blkthr (
  int nx,
  int ny,
  int maxBlocks,
  int maxThreads,
  int &blocksx,
  int &blocksy,
  int &threads
  );

void get_dotwder_blkthr (
  int nx,
  int maxBlocks,
  int maxThreads,
  int &blocksx,
  int &blocksy,
  int &threads
  );

template <class T, unsigned int blockSize, bool nIsPow2>
__global__ void diffvsummsq (
    T *g_ivdata,
    T *g_imdata,
    T *g_odata,
    unsigned int m,
    unsigned int n
    );

#endif
