#include "mex.h"
#include "ggpr.h"
#include "sharedmem.cuh"

#ifdef __DEVICE_EMULATION__
#define EMUSYNC __syncthreads()
#else
#define EMUSYNC
#endif

/* 
 *  kernel copies diagonal elements form matrix S to vector D
 */
__global__ void copyVecDiag(real* D, real* S, size_t pitch, int M)
{
  size_t i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
  real *R;

  R = (real *)((char*)S + i * pitch);
  if (i < M )
    D[i] = log(R[i]);
}

/* 
 * Copy a vector allocated on the device.
 */
__global__ void copyVec ( real* D, real* S, size_t M )
{
  size_t i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
  if ( i < M )
    D[i] = S[i];
}

/**
 * Parallel prefix sum (based on CUDA example).
 * 
 * Can be actually still improved, see other versions below.
 * For CUDA > 2.x it needs `volatile` for the shared memory
 * block.
 */
__global__ void reduce0(real *g_idata, real *g_odata) {
  __shared__ volatile real sdata[BLOCKSIZE];
  // each thread loads one element from global to shared mem
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  sdata[tid] = g_idata[i] + g_idata[i+blockDim.x];
  __syncthreads();

  // do reduction in shared mem
  for (unsigned int s=blockDim.x/2; s>32; s>>=1) {
    if (tid < s) {
      sdata[tid] += sdata[tid + s];
    }
    __syncthreads();
  }
  // last 6 iters are unrolled
  if (tid < 32)
    {
      sdata[tid] += sdata[tid + 32];
      sdata[tid] += sdata[tid + 16];
      sdata[tid] += sdata[tid + 8];
      sdata[tid] += sdata[tid + 4];
      sdata[tid] += sdata[tid + 2];
      sdata[tid] += sdata[tid + 1];
    }
  
  // write result for this block to global mem
  if (tid == 0) g_odata[blockIdx.x] = sdata[0];
}

/*
 * parallel prefix sum (based on CUDA example). For matrices
 */
__global__ void reduce0M(real *g_idata, real *g_odata) {
  __shared__ volatile real sdata[BLOCKSIZE];
  // each thread loads one element from global to shared mem
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  sdata[tid] = g_idata[i] + g_idata[i+blockDim.x];
  __syncthreads();

  // do reduction in shared mem
  for (unsigned int s=blockDim.x/2; s>32; s>>=1) {
    if (tid < s) {
      sdata[tid] += sdata[tid + s];
    }
    __syncthreads();
  }
  // last 6 iters are unrolled
  if (tid < 32)
    {
      sdata[tid] += sdata[tid + 32];
      sdata[tid] += sdata[tid + 16];
      sdata[tid] += sdata[tid + 8];
      sdata[tid] += sdata[tid + 4];
      sdata[tid] += sdata[tid + 2];
      sdata[tid] += sdata[tid + 1];
    }
  
  // write result for this block to global mem
  if (tid == 0) g_odata[blockIdx.x] += sdata[0];
}


/*
 * init matrix 
 */
__global__ void eye(real *A, size_t pitch, size_t M)
{
  size_t i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
  real *R;
  
  R = (real *)((char*)A + i * pitch);
  
  if( i < M )
    {
      R[i] = 1.0;
    }
}

/*
 * Kernel calculates v = v.*v;
 */
__global__ void  MatMul(real *A, size_t pitchA, real *B, size_t pitchB, size_t M)
{
  int i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
  int j = __umul24(blockIdx.y, blockDim.y) + threadIdx.y;

  real *Ar;
  real *Br;

  Ar = (real*) ((char *) A + i * pitchA);
  Br = (real*) ((char *) B + i * pitchB);
  if (i < M && j < M)
    {
      Ar[j] *= Br[j];
    }
}

/* 
 * Initialization of Kss vector 
 */
 __global__ void calcKss( real *Kss, size_t Nin, real sf2, real s2)
 {

   int i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;

   if(i < Nin)
     {
       Kss[i] = sf2 + s2;
     }
 }

/* 
 * This kernel creates the covar. matrix  (covSEard + covNoise)
 */
__global__ void calcTestCovarMatrixBM( real *Kstar, size_t pitchKstar,	\
				     real *X, size_t pitchx,		\
				     real *Xstar, size_t pitchXstar,	\
				     size_t N, size_t M,		\
				     real s2,				\
				     real sf2,				\
				     real *ell,				\
				     size_t NinDim			\
				     )
{
  //int i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x; /* col */
  //int j = __umul24(blockIdx.y, blockDim.y) + threadIdx.y;
  int i = blockIdx.x*blockDim.x	 + threadIdx.x; /* col */
  int j = blockIdx.y*blockDim.y + threadIdx.y;
  int k;
  real dist;
  real *R;
  real *Xr;
  real *Xstarr;
  real t;

  /* TODO: Move the X and Xstar element block into the shared memory
     and compute the `dist` by using SHM data. */

  if ( i < N && j < M )
  {
    R = (real *)((char*)Kstar + i * pitchKstar);
    dist = 0.0;
    for ( k = 0; k < NinDim; k++ )
    {
      Xr = (real *)((char*)X + k * pitchx);
      Xstarr = (real *)((char*)Xstar + k * pitchXstar);
      t = (Xr[j]-Xstarr[i])/ell[k];
      dist += t*t;
    }
    R[j] = sf2*exp(-dist/2);
  }
}

/* 
 * This kernel creates the covar. matrix  (covSEard + covNoise)
 */
__global__ void calcTestCovarMatrixJP
(
  real   *Kstar, size_t pitchKstar,
  real   *X,     size_t pitchX,
  size_t  M,
  real   *Xstar, size_t pitchXstar,
  size_t  Mtd,
  real    s2,
  real    sf2,
  real   *ell,
  size_t  NinDim
  )
{
  /* Matrices are in column-major format. Variables `ri` and `ci`
     hold the row and column indices of the matrix element that
     this particular thread is responsible for. */
  unsigned int ri = blockIdx.x * blockDim.x + threadIdx.x; 
  unsigned int ci = blockIdx.y * blockDim.y + threadIdx.y;
  /* Half of the block height is used to split reading of X and
     Xstar into shared memory. */
  unsigned int halfHeight = blockDim.y/2;
  /* Index of the block row relative to the lower half. Negative
     for blocks that process X, zero or positive for blocks that
     process Xstar. */
  int halfBlockRow = threadIdx.y - halfHeight;
  /* The shared memory blocks of sX and sXstar reside in the same
     block of dynamically allocated shared memory. */
  SharedMemory<real> smem;
  real *sX     = smem.getPointer();
  real *sXstar = sX + blockDim.x*NinDim;

  /* Move the X and Xstar element block into the shared memory
     and compute the `dist` by using SHM data.
     Half of the threads fetches X, the other half fetches Xstar.
     This could be further optimised. */
  if ( halfBlockRow >= 0 )
  {
    /* "Lower" half of the block fetches Xstar.
       
       Xstar is a column-major matrix of `Mtd` rows and `NinDim`
       columns. In the case that `NinDim` is greater than `blockDim.y/2`
       we will have to repeat the reading (not implemented yet).
       
       The column being processed by this thread is equal to the 
       value of `halfBlockRow`, those kernels that correspond to
       columns above `NinDim` do not do anything. */
    if ( halfBlockRow < NinDim && ci < Mtd )
    {
      real *XsCol = (real *)((char*)Xstar + halfBlockRow * pitchXstar);
      unsigned int si = halfBlockRow + NinDim*threadIdx.x;
      /* This should be a coalesced read. May result in shared memory
         bank conflict, though (not checked yet). */
      sXstar[si] = XsCol[ci];
    }
  }
  else
  {
    /* "Upper" half of the block fetches X.
       
       X is a column-major matrix of `M` rows and `NinDim` columns. In
       the case that `NinDim` is greater than `blockDim.y/2` we will
       have to repeat the reading (not implemented yet).
       
       The column being processed by this thread is equal to the 
       value of `threadIdx.y`, those kernels that correspond to columns
       above `NinDim` do not do anything. */
    if ( threadIdx.y < NinDim && ri < M )
    {
      real *XCol = (real *)((char*)X + threadIdx.y * pitchX);
      unsigned int si = threadIdx.y + NinDim*threadIdx.x;
      /* This should be a coalesced read. May result in shared memory
         bank conflict, though (not checked yet). */
      sX[si] = XCol[ri];
    }
  }

  __syncthreads();

  /* Do not compute out-of-bounds elements. */
  if ( ri < M && ci < Mtd )
  {
    /* Colum with result. */
    real *res = (real *)((char*)Kstar + ci * pitchKstar);
    /* Compute vector distance for the two vectors stored in the shared
       memory. The cell coordinate for the result is
       <threadIdx.x,threadIdx.y> where threadIdx.x corresponds to the
       row of X and threadIdx.y corresponds to the row of Xstar. */
    real dist = 0.0;
    size_t k;
    for ( k = 0; k < NinDim; k++ )
    {
      real t = (sX[threadIdx.x*NinDim+k]-sXstar[threadIdx.y*NinDim+k])/ell[k];
      dist += t*t;
    }
    res[ri] = sf2*exp(-dist/2);
  }
}

/*
 * Kernel calculates v = v.*v;
 */ 
__global__ void  squareElements( real *C, size_t pitchC, size_t M, size_t N )
{
  int i = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
  int j = __umul24(blockIdx.y, blockDim.y) + threadIdx.y;

  real *Cr;
  real x;

  Cr = (real*) ((char *) C + i * pitchC);
  if (i < N && j < M)
    {
      x = Cr[j];
      Cr[j] = __dmul_rn(x,x);
      /*Cr[j] = Cr[j]*Cr[j];*/
    }
}


/**
 * Return next value of 2^i for the given number.
 */
unsigned int nextPow2( unsigned int x )
{
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}

bool isPow2(unsigned int x)
{
    return ((x&(x-1))==0);
}

/**
 *  Sum elements of a double vector.
 *  Based on CUDA parallel prefix sum example, kernel 6.
 *
 *  @param g_data device address of the summed vector
 *  @param m      length of the vector
 */
template <unsigned int blockSize, bool nIsPow2>
__global__ void gpusum (
    double *g_data,
    unsigned int n
    )
{
    SharedMemory<double> smem;
    double *sdata = smem.getPointer();

    // in case of emulating the device remember templating values
#ifdef __DEVICE_EMULATION__
    unsigned int _blockSize = blockSize;
    bool         _mIsPow2   = mIsPow2;
#endif
    unsigned int tid = threadIdx.x;
    unsigned int i   = blockIdx.x * (blockSize*2) + tid;
    unsigned int gridSize = blockSize*2 * gridDim.x;
	// sum in a register is faster than in shared memory
	register double tsum = 0.0;

	// Perform the first level of reduction, reading from global memory and
    // writing to shared memory. We reduce multiple elements per thread.
    // The number is determined by the number of active thread blocks
    // (via gridDim). More blocks will result in a larger gridSize and
    // therefore fewer elements per thread.
    while (i < n)
    {         
        tsum += g_data[i];
        // ensure we don't read out of bounds -- this is optimized away for powerOf2 sized arrays
        if (nIsPow2 || i + blockSize < n) 
            tsum += g_data[i+blockSize];  
        i += gridSize;
    } 
    sdata[tid] = tsum;
    __syncthreads();

    // do reduction in shared mem
    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
#ifndef __DEVICE_EMULATION__
    if (tid < 32)
#endif
    {
        /* What follows should be warp synchronous, but due to compiler
		   optimalisations when compiling for Fermi, the original CUDA 2.3
		   example will not work. We therefore chage the memory location
		   to `volatile`. */
		volatile double *warpMem = sdata;
        if (blockSize >=  64) { warpMem[tid] += warpMem[tid + 32]; EMUSYNC; }
        if (blockSize >=  32) { warpMem[tid] += warpMem[tid + 16]; EMUSYNC; }
        if (blockSize >=  16) { warpMem[tid] += warpMem[tid +  8]; EMUSYNC; }
        if (blockSize >=   8) { warpMem[tid] += warpMem[tid +  4]; EMUSYNC; }
        if (blockSize >=   4) { warpMem[tid] += warpMem[tid +  2]; EMUSYNC; }
        if (blockSize >=   2) { warpMem[tid] += warpMem[tid +  1]; EMUSYNC; }
    }
    
    // write result for this block to global mem 
    if (tid == 0) g_data[blockIdx.x] = sdata[0];
}

/**
 *  Computes a sum of dot products.
 *  Used to evaulate the partial derivatives when training the GP according to
 *
 *      d = sum(sum(W.*Der));
 *
 *  (see `gpr_training.m` for details).
 *  Based on CUDA parallel prefix sum example, tweaked to our needs. Matrices
 *  are in column-major format (incremental index changes traverses rows in
 *  a column) and allocated with `cudaMallocPitch()` to enable data coalescing.
 *
 *  @param q_iwdata device address of the input matrix `w` of size `m`x`m`
 *  @param q_iddata device address of the input matrix `v` of size `m`x`m`
 *  @param q_iodata device address of the output scalar value `out2`
 *  @param m        dimension of the matrices
 *  @param pitch  pitch of the columns
 */
template <unsigned int blockSize, bool mIsPow2>
__global__ void dotwder (
    double *g_iwdata,
    double *g_iddata,
    double *g_odata,
    unsigned int m,
    size_t pitch
    )
{
    SharedMemory<double> smem;
    double *sdata = smem.getPointer();

    // in case of emulating the device remember templating values
#ifdef __DEVICE_EMULATION__
    unsigned int _blockSize = blockSize;
    bool         _mIsPow2   = mIsPow2;
#endif
    unsigned int tid = threadIdx.x;
    // what is the column we are processing (Matlab column-major format,
    // in C format this would be the row that is being processed)
    unsigned int col  = gridDim.y*blockIdx.x + blockIdx.y;
    // position of the element in the current column that we are going
    // to process as the first one, maximum index is `m`
    unsigned int i    = tid;
    // we will process one row in a thread block
    unsigned int gridSize = 2*blockSize;
	// sum in a register is faster than in shared memory
	register double tsum = 0.0;

    // Update the `g_iwdata` and `g_iddata` pointers so that they point to the
    // beginning of the current column.
    g_iwdata = (double*)((char*)g_iwdata + col*pitch);
    g_iddata = (double*)((char*)g_iddata + col*pitch);

    // Perform first level of reduction, (hopefully coalesced) reading from global
    // memory, writing to shared memory. We reduce multiple elements per thread.
    // The number of reduced elements is determined by the number of active thread
    // blocks (via gridDim).  More blocks will result in a larger gridSize and
    // therefore fewer elements per thread being reduced.
    while ( i < m )
    {         
        tsum += g_iwdata[i] * g_iddata[i];
        // ensure we don't read out of bounds -- this is optimized away for powerOf2 sized arrays
        if ( mIsPow2 || i + blockSize < m )
        {
            tsum += g_iwdata[i+blockSize] * g_iddata[i+blockSize];
        }
        i += gridSize;
    } 
    sdata[tid] = tsum;
	__syncthreads();

    // do reduction in shared memory
    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
#ifndef __DEVICE_EMULATION__
    if (tid < 32)
#endif
    {
        /* What follows should be warp synchronous, but due to compiler
		   optimalisations when compiling for Fermi, the original CUDA 2.3
		   example will not work. We therefore chage the memory location
		   to `volatile`. */
		volatile double *warpMem = sdata;
        if (blockSize >=  64) { warpMem[tid] += warpMem[tid + 32]; EMUSYNC; }
        if (blockSize >=  32) { warpMem[tid] += warpMem[tid + 16]; EMUSYNC; }
        if (blockSize >=  16) { warpMem[tid] += warpMem[tid +  8]; EMUSYNC; }
        if (blockSize >=   8) { warpMem[tid] += warpMem[tid +  4]; EMUSYNC; }
        if (blockSize >=   4) { warpMem[tid] += warpMem[tid +  2]; EMUSYNC; }
        if (blockSize >=   2) { warpMem[tid] += warpMem[tid +  1]; EMUSYNC; }
    }
    
    // add result for this block to the global sum
    if (tid == 0)
      g_odata[col] = sdata[0];
}

/**
 *  Computes vector minus a sum of summed squared elements of a matrix.
 *  Used as the final step of GP simulation where out2 has to be computed
 *  according to
 *
 *      out2 = k - sum(v.*v)';
 *
 *  (see `grp_simul.m` for details).
 *  Based on CUDA parallel prefix sum example, tweaked to our needs. Matrices
 *  are in column-major format (incremental index changes traverses rows in
 *  a column) and allocated with `cudaMallocPitch()` to enable data coalescing.
 *
 *  The limit on block size of 65535x65535 blocks makes it necessary to split
 *  the block numbering of larger matrices so that x-dimension of the grid
 *  correspond to lower bits of column index and the y-dimension corresponds
 *  to higher bits of the index so that gridDim.x<65535 and gridDim.y<65535.
 *
 *  @param q_ivdata device address of the input column vector `k` of lenght `m`
 *  @param q_imdata device address of the input matrix `v` of size `m`x`n`
 *  @param q_iodata device address of the output vector `out2` of length `m`
 *  @param m        number of rows of the matrix `v` and vector lengths
 *  @param n        number of columns of the matrix `v`
 *  @param v_pitch  pitch of the column space in `v` in bytes
 */
template <unsigned int blockSize, bool mIsPow2>
__global__ void diffvsummsq (
    double *g_ikdata,
    double *g_ivdata,
    double *g_odata,
    unsigned int m,
    unsigned int n,
    size_t v_pitch
    )
{
    SharedMemory<double> smem;
    double *sdata = smem.getPointer();

    // in case of emulating the device remember templating values
#ifdef __DEVICE_EMULATION__
    unsigned int _blockSize = blockSize;
    bool         _mIsPow2   = mIsPow2;
#endif
    unsigned int tid = threadIdx.x;
    // what is the column we are processing (Matlab column-major format,
    // in C format this would be the row that is being processed)
    unsigned int col  = gridDim.y*blockIdx.x + blockIdx.y;
    // position of the element in the current column that we are going
    // to process as the first one, maximum index is `m`
    unsigned int i    = tid;
    // we will process one row in a thread block
    unsigned int gridSize = 2*blockSize;
	// sum in a register is faster than in shared memory
	register double tsum = 0.0;

    // Update the `g_ivdata` pointer so that it points to the beginning of
    // the current column.
    g_ivdata = (double*)((char*)g_ivdata + col*v_pitch);

	// Perform first level of reduction, (hopefully coalesced) reading from global
    // memory, writing to shared memory. We reduce multiple elements per thread.
    // The number of reduced elements is determined by the number of active thread
    // blocks (via gridDim).  More blocks will result in a larger gridSize and
    // therefore fewer elements per thread being reduced.
    while ( i < m )
    {         
        double elem = g_ivdata[i];
        tsum += elem*elem;
        // ensure we don't read out of bounds -- this is optimized away for powerOf2 sized arrays
        if ( mIsPow2 || i + blockSize < m )
        {
            elem = g_ivdata[i+blockSize];
            tsum += elem*elem;
        }
        i += gridSize;
    }
	sdata[tid] = tsum;
    __syncthreads();

    // do reduction in shared mem
    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
#ifndef __DEVICE_EMULATION__
    if (tid < 32)
#endif
    {
        /* What follows should be warp synchronous, but due to compiler
		   optimalisations when compiling for Fermi, the original CUDA 2.3
		   example will not work. We therefore chage the memory location
		   to `volatile`. */
		volatile double *warpMem = sdata;
		if (blockSize >=  64) { warpMem[tid] += warpMem[tid + 32]; EMUSYNC; }
        if (blockSize >=  32) { warpMem[tid] += warpMem[tid + 16]; EMUSYNC; }
        if (blockSize >=  16) { warpMem[tid] += warpMem[tid +  8]; EMUSYNC; }
        if (blockSize >=   8) { warpMem[tid] += warpMem[tid +  4]; EMUSYNC; }
        if (blockSize >=   4) { warpMem[tid] += warpMem[tid +  2]; EMUSYNC; }
        if (blockSize >=   2) { warpMem[tid] += warpMem[tid +  1]; EMUSYNC; }
    }
    
    // write result for this block to global mem 
    if (tid == 0)
      g_odata[col] = g_ikdata[col] - sdata[0];
      //g_odata[col] = sdata[0];
	  //g_odata[0] = sdata[0];
}

template <class T>
__global__ void sample ( T *data)
{
  T val;
  val = *data;
  val = val*2;
}

/**
 * Wrapper function for launching `diffvsummsq` kernel.
 */
void gpusum_wrapper (
    int size,
    int threads,
    int blocks,
    double *d_data
    )
{
    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(blocks, 1, 1);
    int smemSize = threads * sizeof(double);

    if (isPow2(size))
    {
        switch (threads)
        {
        case 512:
            gpusum<512, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 256:
            gpusum<256, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 128:
            gpusum<128, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 64:
            gpusum< 64, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 32:
            gpusum< 32, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 16:
            gpusum< 16, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  8:
            gpusum<  8, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  4:
            gpusum<  4, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  2:
            gpusum<  2, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  1:
            gpusum<  1, true><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        }
    }
    else
    {
        switch (threads)
        {
        case 512:
            gpusum<512, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 256:
            gpusum<256, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 128:
            gpusum<128, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 64:
            gpusum< 64, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 32:
            gpusum< 32, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case 16:
            gpusum< 16, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  8:
            gpusum<  8, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  4:
            gpusum<  4, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  2:
            gpusum<  2, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        case  1:
            gpusum<  1, false><<< dimGrid, dimBlock, smemSize >>>(d_data, size); break;
        }
    }
}

/**
 * Wrapper function for launching `dotwder` kernel.
 */
void dotwder_wrapper (
    int size,
    int threads,
    int blocksX,
    int blocksY, 
    double *d_iwdata,
    double *d_iddata,
    size_t  pitch,
    double *d_odata
    )
{
    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(blocksX, blocksY, 1);
    int smemSize = threads * sizeof(double);

#ifdef DEBUG
    mexPrintf ( "    dotwder_wrapper: block is (%d,%d,%d) threads,", dimBlock.x, dimBlock.y, dimBlock.z );
    mexPrintf ( " grid  is (%d,%d,%d) blocks,", dimGrid.x, dimGrid.y, dimGrid.z );
    mexPrintf ( " shared memory size is %d bytes\n", smemSize );
#endif

    if (isPow2(size))
    {
        switch (threads)
        {
        case 512:
            dotwder<512, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 256:
            dotwder<256, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 128:
            dotwder<128, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 64:
            dotwder< 64, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 32:
            dotwder< 32, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 16:
            dotwder< 16, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  8:
            dotwder<  8, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  4:
            dotwder<  4, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  2:
            dotwder<  2, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  1:
            dotwder<  1, true><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        }
    }
    else
    {
        switch (threads)
        {
        case 512:
            dotwder<512, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 256:
            dotwder<256, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 128:
            dotwder<128, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 64:
            dotwder< 64, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 32:
            dotwder< 32, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case 16:
            dotwder< 16, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  8:
            dotwder<  8, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  4:
            dotwder<  4, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  2:
            dotwder<  2, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        case  1:
            dotwder<  1, false><<< dimGrid, dimBlock, smemSize >>>(d_iwdata, d_iddata, d_odata, size, pitch); break;
        }
    }
}

/**
 * Wrapper function for launching `diffvsummsq` kernel.
 */
void diffvsummsq_wrapper (
    int sizex, // M (number of rows in `v`
    int sizey, // Mtd (number of columns in `v` and rows in `Kss`)
    int threads,
    int blocksX,
    int blocksY, 
    double *d_ivdata,
    double *d_imdata,
    size_t pitch,
    double *d_odata
    )
{
    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(blocksX, blocksY, 1);
    int smemSize = threads * sizeof(double);

#ifdef DEBUG
    mexPrintf ( "    diffvsummsq_wrapper: block is (%d,%d,%d) threads,", dimBlock.x, dimBlock.y, dimBlock.z );
    mexPrintf ( " grid  is (%d,%d,%d) blocks,", dimGrid.x, dimGrid.y, dimGrid.z );
    mexPrintf ( " shared memory size is %d bytes\n", smemSize );
    mexPrintf ( "    template: <blockSize=%d,mIsPow2=%d>\n", threads, isPow2(sizex));
#endif

    if (isPow2(sizex))
    {
        switch (threads)
        {
        case 512:
            diffvsummsq<512, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 256:
            diffvsummsq<256, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 128:
            diffvsummsq<128, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 64:
            diffvsummsq< 64, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 32:
            diffvsummsq< 32, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 16:
            diffvsummsq< 16, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  8:
            diffvsummsq<  8, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  4:
            diffvsummsq<  4, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  2:
            diffvsummsq<  2, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  1:
            diffvsummsq<  1, true><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        }
    }
    else
    {
        switch (threads)
        {
        case 512:
            diffvsummsq<512, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 256:
            diffvsummsq<256, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 128:
            diffvsummsq<128, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 64:
            diffvsummsq< 64, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 32:
            diffvsummsq< 32, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case 16:
            diffvsummsq< 16, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  8:
            diffvsummsq<  8, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  4:
            diffvsummsq<  4, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  2:
            diffvsummsq<  2, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        case  1:
            diffvsummsq<  1, false><<< dimGrid, dimBlock, smemSize >>>(d_ivdata, d_imdata, d_odata, sizex, sizey, pitch); break;
        }
    }
}

#ifdef NEVER
extern "C"
void diffvsummsq_wrapper (
    int sizex,
    int sizey,
    int threads,
    int blocksX,
    int blocksY, 
    double *d_ivdata,
    double *d_imdata,
    double *d_odata
    )
{
  //sample<double>(d_ivdata);
  //diffvsummsq_tpl<double>(sizex, sizey, threads, blocksX, blocksY, d_ivdata, d_imdata, d_odata);
}
#endif

/**
 * Compute the number of threads and blocks to use for the gpusum kernel.
 * Adopted from reduce demo - we set threads / block to the minimum of maxThreads
 * and n/2 and we observe the maximum specified number of blocks, because each
 * thread in the kernel can process a variable number of elements.
 */
void get_gpusum_blkthr (
  int n,
  int maxBlocks,
  int maxThreads,
  int &blocks,
  int &threads
  )
{
  threads = (n < maxThreads*2) ? nextPow2((n + 1)/ 2) : maxThreads;
  blocks = (n + (threads * 2 - 1)) / (threads * 2);
  blocks = min(maxBlocks, blocks);
}

/**
 * Determine the number of blocks and threads for `dotwder` kernel.
 */
void get_dotwder_blkthr (
  int nx,
  int maxBlocks,
  int maxThreads,
  int &blocksx,
  int &blocksy,
  int &threads)
{
  // For smaller matrix dimensions we will not use the maximum of
  // threads per block, but we will rather limit the number to the
  // newx power of two
  threads = (nx < maxThreads*2) ? nextPow2((nx + 1)/ 2) : maxThreads;
  
  // The matrices are `nx` x `nx` elements, every block is supposed to
  // process one column of the length `nx`. Unless the matrix dimensions
  // exceed 65535 (which is the maximum number of blocks in one dimension
  // of a grid), we are okay by setting the grid to (1,nx,1). For higher
  // dimensions, the matrix will be split into `blocksx` parts of
  // `blocksy` columns.
  blocksx = 1;
  blocksy = nx;
  while ( blocksy > 65535 )
  {
    blocksx = blocksx<<1;
    blocksy = blocksy>>1;
  }
}

/**
 * Determine the number of blocks and threads for `diffvsummsq` kernel.
 * 
 * @params nx The number of rows in matrix v (M)
 * @params ny The length of Kss vector (Mtd)
 */
void get_diffvsummsq_blkthr (
  int nx,
  int ny,
  int maxBlocks,
  int maxThreads,
  int &blocksx,
  int &blocksy,
  int &threads)
{
  // For smaller matrix dimensions we will not use the maximum of
  // threads per block, but we will rather limit the number to the
  // newx power of two
  threads = (nx < maxThreads*2) ? nextPow2((nx + 1)/ 2) : maxThreads;

  // The matrix `v` is `nx` x `ny` elements and every block is supposed
  // to process one column of the length `nx` and subtract the result from
  // the corresponding `Kss` element where `ny` is the lenght of `Kss`.
  // Unless the `ny` dimension exceeds 65535 (which is the maximum number
  // of blocks in one dimension of a grid), we may set the grid to (1,ny,1).
  // For higher dimensions, the matrix `v` will be split into `blocksx` parts
  // of `blocksy` columns.
  blocksx = 1;
  blocksy = ny;
  while ( blocksy > 65535 )
  {
    blocksx = blocksx<<1;
    blocksy = blocksy>>1;
  }
}
